"""
py2json utils functions and other helpers
"""
from i2.routing_forest import *
from i2.footprints import *
from dataclasses import dataclass


@dataclass
class LiteralVal:
    obj: object


def missing_args_func(func_to_kwargs=None, ignore=None):
    """Returns a function that returns a set of missing arg names.
    Missing means that that the arg is required for a func (has no default)
    and hasn't been found by the func_to_kwargs policy.

    The returned function can be used to diagnose the coverage of the func_to_kwargs policy,
    or as a filter to find those functions that are not covered by the policy.

    :param func_to_kwargs: Callable that returns valid kwargs for a given func.
    :param ignore: If not None, should be an iterable to names not to check
    :return: A missing_args_func that returns a set of arg names that are missing.

    >>> from collections import namedtuple
    >>> assert missing_args_func()(namedtuple) == {'field_names', 'typename'}
    >>> func_to_kwargs = lambda f: {namedtuple: {'typename': 'Unspecified'}}.get(f, {})
    >>>
    >>> missing_args = missing_args_func(func_to_kwargs)
    >>> missing_args(namedtuple)
    {'field_names'}
    >>> def foo(x=1, y=2): ...  # defaults cover all arguments
    >>> assert list(filter(missing_args, (namedtuple, foo))) == [namedtuple]
    """

    def missing_args_func_(func):
        missing_args = set(Sig(func).without_defaults.parameters) - set(ignore or ())
        if func_to_kwargs is not None:
            missing_args -= func_to_kwargs(func).keys()
        return missing_args

    return missing_args_func_


def mk_func_to_kwargs_from_a_val_for_argname_map(val_for_argname=None):
    """Returns a function func_to_kwargs that returns kwargs for a given callable func.
    The intent being that these kwargs can be used as valid inputs of func as such:
    ```
        func(**func_to_kwargs)
    ```

    Does so by taking the intersection of those arguments of the func that don't have defaults
    and the input val_for_argname mapping.

    Note that if no val_for_argname is given, or non matches the default-less arguments of func,
    then {} is returned.

    >>> val_for_argname = {'typename': 'Unspecified', 'x': 0}
    >>> func_to_kwargs = mk_func_to_kwargs_from_a_val_for_argname_map(val_for_argname)
    >>> missing_args = missing_args_func(func_to_kwargs)
    >>>
    >>> from collections import namedtuple
    >>> def foo(typename, x, y=2): ...
    >>> def bar(x, z=None): ...
    >>> assert missing_args(namedtuple) == {'field_names'}
    >>> assert missing_args(foo) == set()
    >>> assert missing_args(bar) == set()
    >>> assert list(filter(missing_args, (namedtuple, foo, bar))) == [namedtuple]
    """
    val_for_argname = val_for_argname or {}

    def func_to_kwargs(func):
        return {
            k: val_for_argname[k]
            for k in val_for_argname.keys() & set(Sig(func).without_defaults)
        }

    return func_to_kwargs


def is_valid_kwargs(func, kwargs):
    """Test if kwargs constitute a valid input for func simply by trying func(**kwargs) out.

    :param func: A callable
    :param kwargs: A dict of keyword arguments
    :return: True if, and only if `func(**kwargs)` doesn't fail, and False if it does raise an Exception.

    >>> def f(a, b=1):
    ...     return a * b
    >>> is_valid_kwargs(f, {'a': 10})
    True
    >>> is_valid_kwargs(f, {'a': 1, 'b': 10})
    True
    >>> is_valid_kwargs(f, {'b': 2, 'c': 4})  # c is not a valid argument name, so...
    False
    >>> is_valid_kwargs(f, {})  # a has no default value, so you need at least that argument, so...
    False
    """
    try:
        func(**kwargs)
        return True
    except Exception as e:
        return False


from inspect import signature
from functools import wraps, reduce
from i2.signatures import Sig
import warnings


def is_not_none(x):
    return x is not None


def ignore_warnings(func):
    """Couldn't make this work"""
    from i2.signatures import Sig

    raise RuntimeError("Couldn't make this work. Don't use!")

    @Sig.from_objs(func, lambda ignore_warnings=True: ...)
    def _func(*args, ignore_warnings=True, **kwargs):
        with warnings.catch_warnings():
            if ignore_warnings:
                warnings.simplefilter('ignore')
            return func(*args, **kwargs)

    return _func


def catch_errors(errors=(Exception,), on_error=lambda e: print(e)):
    if not callable(on_error):
        on_error_val = on_error
        on_error = lambda e: on_error_val
    else:
        nargs = len(signature(on_error).parameters)
        if nargs > 1:
            raise ValueError(
                f'on_error should be a value or callable with 0 or 1 arguments'
            )
        elif nargs == 0:
            on_error_func = on_error
            on_error = lambda e: on_error_func()

    def wrap_func(func):
        @wraps(func)
        def func_with_errors_caught(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except errors as e:
                return on_error(e)

        return func_with_errors_caught

    return wrap_func


class Nones:
    """
    >>> x, y, z = Nones(3)
    >>> x, y, z
    (None, None, None)
    >>> bool(Nones(3))
    False
    """

    def __init__(self, n_items: int):
        self.n_items = n_items

    def __iter__(self):
        return (None for _ in range(self.n_items))

    def __bool__(self):
        return False


def partial_positionals(func, fix_args, **fix_kwargs):
    """Like functools.partial, but with positionals as well"""

    def wrapper(*args, **kwargs):
        arg = iter(args)
        return func(
            *(
                fix_args[i] if i in fix_args else next(arg)
                for i in range(len(args) + len(fix_args))
            ),
            **{**fix_kwargs, **kwargs},
        )

    return wrapper


def is_types_spec(types) -> bool:
    """Returns True iff input types is a type or an iterable of types"""
    if isinstance(types, type):
        types = (types,)
    try:
        return len(types) > 0 and all(isinstance(x, type) for x in types)
    except Exception:
        return False


def mk_isinstance_cond(types) -> Callable[[Any], bool]:
    """Makes a boolean function that verifies if objects are of a target type (or types)"""

    assert is_types_spec(types), f'types need to be a single or an iterable of types'

    def isinstance_of_target_types(obj):
        return isinstance(obj, types)

    return isinstance_of_target_types


def mk_scan_mapper(condition_map, dflt=None):
    """Make function implementing an if/elif/.../else logic from a {bool_func: x, ...} map"""

    def scan_mapping(x):
        for condition, then in condition_map.items():
            if condition(x):
                return then
        return dflt

    return scan_mapping


def types_to_cond(types_map):
    """Convert a {type(s): x, ...} map into a {is_of_that_type: x, ...} map"""
    return {mk_isinstance_cond(types): x for types, x in types_map.items()}


def types_map_to_scan_mapper(types_map, dflt=None):
    """Make function implementing an if/elif/.../else logic from a {type(s): x, ...} map.
    The returned mapper will be such that `mapper(obj)` will return a value x
    according to the first `isinstance(obj, types)` check that is found in {types: x, ...}
    types_map.

    >>> mapper = types_map_to_scan_mapper({dict: 'a dict!', (list, tuple): 'list-like'},
    ...     dflt='nothing found')
    >>> mapper({'a': 'dict'})
    'a dict!'
    >>> mapper((1, 2, 3))
    'list-like'
    >>> mapper(['a', 'list'])
    'list-like'
    >>> mapper(lambda x: x)  # a function: No match for that!
    'nothing found'
    >>> mapper(mapper)
    'nothing found'
    """
    return mk_scan_mapper(types_to_cond(types_map), dflt)


def compose(*functions):
    """Make a function that is the composition of the input functions"""
    return reduce(lambda f, g: lambda x: f(g(x)), functions, lambda x: x)



"""
Navigation and serialization of objects using glom
"""
from glom import glom, Spec
import importlib


def mk_serializer_and_deserializer(spec, mk_inv_spec=None):
    if not isinstance(spec, Spec):
        spec = Spec(spec)

    def serialize(o):
        return glom(o, spec)

    if mk_inv_spec is None:
        return serialize
    else:

        def deserialize(o):
            return glom(o, mk_inv_spec(o))

        return serialize, deserialize


specs_for_kind = {
    'jdict_methods': {
        'description': 'For objects that have from_jdict and to_jdict methods',
        'spec': {
            'module': '__class__.__module__',
            'name': '__class__.__name__',
            'attr': 'from_jdict.__name__',
            'kwargs': lambda x: x.to_jdict(),
        },
        'mk_inv_spec': lambda d: (
            lambda x: importlib.import_module(d['module']),
            lambda x: getattr(x, d['name']),
            lambda x: getattr(x, d['attr']),
            lambda x: x.__call__(d['kwargs']),
        ),
    }
}


def mk_serializer_and_deserializer_for_kind(kind='jdict_methods'):
    specs_dict = specs_for_kind[kind]
    return mk_serializer_and_deserializer(specs_dict['spec'], specs_dict['mk_inv_spec'])



"""
f
"""
from anytree import Node, AnyNode, RenderTree, ContStyle, NodeMixin
from warnings import warn

warn(
    'Keeping around until final kv_walk based version is done',
    PendingDeprecationWarning,
)

NodeMixin.separator = '.'

DFLT_ON_ERROR = 'warn'


def handle_error(e, on_error=DFLT_ON_ERROR):
    if on_error == 'warn':
        warn(str(e))
    elif on_error == 'skip':
        pass
    elif on_error == 'raise':
        raise
    else:
        raise ValueError(f'Unknown on_error value: {on_error}')


def get_attr_and_vals(obj, on_error=DFLT_ON_ERROR):
    for k in dir(obj):
        try:
            yield k, getattr(obj, k)
        except Exception as e:
            handle_error(e, on_error)


def attr_and_vals_dict(obj, filt=None):
    return {k: v for k, v in filter(filt, get_attr_and_vals(obj))}


# Filters
def value_not_callable(kv):
    return not callable(kv[1])


class AttrTreeImporter:
    _base_types = (int, float, str, bytes, bool, list)

    def __init__(self, val_types=(), max_levels=99, on_error=DFLT_ON_ERROR):
        self.val_types = tuple(set(val_types).union(self._base_types))
        self.on_error = on_error
        self.max_levels = max_levels

    def attr_and_vals(self, obj, level=0):
        if level < self.max_levels:
            for k, v in get_attr_and_vals(obj):
                if isinstance(v, self.val_types):
                    yield k, v
                else:
                    yield k, self.dict_of_obj(v, level + 1)

    def dict_of_obj(self, obj, level=0):
        return dict(self.attr_and_vals(obj, level))

    def tree_of_dict(self, d, parent=None, root_name='root'):
        if parent is None:
            parent = Node(root_name)
        for k, v in d.items():
            if isinstance(v, dict):
                n = Node(k, parent=parent)
                self.tree_of_dict(v, parent=n)
            else:
                n = Node(k, parent=parent, val=v)
        return parent

    def tree_of_obj(self, obj, root_name=None):
        if root_name is None:
            if hasattr(obj, '__name__'):
                root_name = obj.__name__
            elif hasattr(obj, '__class__'):
                root_name = obj.__class__.__name__
            else:
                root_name = 'root'
        return self.tree_of_dict(self.dict_of_obj(obj), root_name=root_name)



"""
f
"""
import inspect


def func_to_jdict(func):
    """
    >>> def multiplier(a, b):
    ...     return a * b
    >>> jdict = func_to_jdict(multiplier)
    >>> assert jdict == {'$py_source_lines': 'def multiplier(a, b):\\n    return a * b\\n'}
    """
    lines = inspect.getsource(func)
    return {'$py_source_lines': lines}


def jdict_to_func(jdict):
    """
    >>> jdict = {'$py_source_lines': 'def multiplier(a, b):\\n    return a * b\\n'}
    >>> deserialized_func = jdict_to_func(jdict)
    >>> deserialized_func(7, 6)
    42
    """
    _locals = {}
    exec(jdict['$py_source_lines'], None, _locals)
    func_name, func_obj = next(iter(_locals.items()))
    return func_obj


def test_this():
    def multiplier(a, b):
        return a * b

    func = multiplier

    jdict = func_to_jdict(func)
    assert jdict == {'$py_source_lines': 'def multiplier(a, b):\n    return a * b\n'}

    deserialized_func = jdict_to_func(jdict)
    assert deserialized_func != func  # not equal, but....
    assert deserialized_func(7, 6) == 42



"""
An attempt at getting a recursive attribute tree
"""
##### Utils #####################################################################################
## To be able to do partial with positionals too


# Explicit version of partial_positionals(isinstance, {1: types})
from py2json.util import mk_isinstance_cond, mk_scan_mapper
import numpy as np

isinstance_mapping = mk_isinstance_cond(np.ndarray)
assert isinstance_mapping([1, 2, 3]) == False
assert isinstance_mapping(np.array([1, 2, 3])) == True


def serialized_attr_dict(obj, serializer, attrs=None):
    attrs = attrs or dir(obj)
    return {a: serializer(getattr(obj, a)) for a in attrs}


class Struct:
    def __init__(self, **kwargs):
        for a, val in kwargs.items():
            setattr(self, a, val)


def deserialize_as_obj(attr_dict, deserializer, cls=Struct):
    obj = cls()
    for k, v in attr_dict.items():
        setattr(obj, k, deserializer(v))
    return obj


##### Use #####################################################################################
import numpy
import pandas
from py2json.fakit import refakit

from py2json.util import is_types_spec
from py2json.fakit import is_valid_fak
from i2.deco import postprocess, preprocess


@postprocess(dict)
def mk_cond_map_from_types_map(types_map):
    for types, serializer in types_map.items():
        if is_types_spec(types):
            types = mk_isinstance_cond(types)
        assert callable(
            types
        ), f'types spec should be a callable at this point: {types}'
        # TODO: Would lead to shorter spec language, but needs "arg injection" of sorts
        # if isinstance(serializer, (dict, tuple, list)):
        #     assert is_valid_fak(serializer), f"Should be a valid fak: {serializer}"
        #     fak_spec = serializer
        #
        #     def serializer(x):
        #         return {'$fak': fak_spec}
        yield types, serializer


def asis(x):
    return x


def mk_serializer_and_deserializer_for_types_map(types_map):
    cond_map = mk_cond_map_from_types_map(types_map)
    scan_mapper = mk_scan_mapper(cond_map, dflt=asis)

    def serializer(obj):
        return scan_mapper(obj)(obj)

    return serializer, refakit


# TODO: much to factor out into a mini-language here
# TODO: See how the specs complexify if we want to use orient='records' kw in DataFrame (de)serialization
type_cond_map = {
    numpy.ndarray: lambda x: {'$fak': ('numpy.array', (numpy.ndarray.tolist(x),))},
    pandas.DataFrame: lambda x: {
        '$fak': {
            'f': 'pandas.DataFrame.from_dict',
            'k': {
                'data': pandas.DataFrame.to_dict(x, orient='index'),
                'orient': 'index',
            },
        }
    },
}

fak_serialize, fak_deserialize = mk_serializer_and_deserializer_for_types_map(
    type_cond_map
)

arr = numpy.array([1, 2, 3])
serialized_arr = fak_serialize(arr)
assert serialized_arr == {'$fak': ('numpy.array', ([1, 2, 3],))}
assert all(fak_deserialize(serialized_arr) == arr)

df = pandas.DataFrame({'foo': [1, 2, 3], 'bar': [10, 20, 30]})
serialized_df = fak_serialize(df)
assert serialized_df == {
    '$fak': {
        'f': 'pandas.DataFrame.from_dict',
        'k': {
            'data': {
                0: {'foo': 1, 'bar': 10},
                1: {'foo': 2, 'bar': 20},
                2: {'foo': 3, 'bar': 30},
            },
            'orient': 'index',
        },
    }
}
assert all(fak_deserialize(serialized_df) == df)


##### Now use with sklearn model! #########################################################


class GenericEstimator(Struct):
    fit = None  # or sklearn will complain


from functools import partial, wraps
from i2.footprints import attrs_used_by_method
import json


# def jsonize_output(func):
#     @wraps(func)
#     def wrapper(x):
#         try:
#             json.dumps()


def mk_serializer_and_deserializer_from_instance_and_methods(
    instance, methods, deserialize_to_cls=None, jsonize=True
):
    cls = instance.__class__
    cls_has_attr = lambda a: hasattr(cls, a)
    instance_has_attr = lambda a: hasattr(instance, a)

    @postprocess(set)
    def get_needed_attrs():
        for method in filter(cls_has_attr, methods):
            target_func = getattr(cls, method)
            for attr in filter(instance_has_attr, attrs_used_by_method(target_func)):
                yield attr

    deserialize_to_cls = deserialize_to_cls or cls
    serializer = partial(
        serialized_attr_dict, serializer=fak_serialize, attrs=get_needed_attrs(),
    )
    deserializer = partial(
        deserialize_as_obj, deserializer=fak_deserialize, cls=deserialize_to_cls,
    )

    if jsonize:
        serializer = postprocess(json.dumps, verbose_error_message=2)(serializer)
        deserializer = preprocess(json.loads)(deserializer)

    return serializer, deserializer


#
#
# my_serializer = partial(serialized_attr_dict, serializer=fak_serialize)
# my_deserializer = partial(deserialize_as_obj, deserializer=fak_deserialize, cls=GenericEstimator)

# make something to serialize
from sklearn.decomposition import PCA
from sklearn.datasets import make_blobs

X, y = make_blobs()
model = PCA().fit(X)

(serialize, deserialize,) = mk_serializer_and_deserializer_from_instance_and_methods(
    model, methods=['predict', 'transform'], deserialize_to_cls=None
)

serialized_model = serialize(model)
deserialized_model = deserialize(serialized_model)

assert isinstance(serialized_model, str)
if hasattr(model, 'transform'):
    target_func = model.__class__.transform
else:
    target_func = model.__class__.predict

assert numpy.allclose(target_func(deserialized_model, X), target_func(model, X))


# and in case you forgot the equivalence:
# assert (model.transform(X) == target_func(model, X)).all()


def mk_serialize_deserialize_pipeline_for_model(
    model, methods=('predict', 'transform'), deserialize_to_cls=None
):
    (
        serialize,
        deserialize,
    ) = mk_serializer_and_deserializer_from_instance_and_methods(
        model, methods=methods, deserialize_to_cls=deserialize_to_cls
    )

    def alt_model(model):
        return deserialize(serialize(model))

    return alt_model


from py2json.inspire.serializing_sklearn_estimators import estimator_test_df

mk_alt_model = lambda model: mk_serialize_deserialize_pipeline_for_model(model)(model)

kwargs_for_behavioral_test_kwargs_for_estimator = dict(
    mk_alt_model=mk_alt_model,  # Returns an alt object by serializing and deserializing the fitted model
    methods=('predict', 'transform',),  # The methods to use to make behavior_funcs,
)


def test_estimators_serialization(estimator_classes=None, ignore_warnings=True):
    return estimator_test_df(
        estimator_classes,
        ignore_warnings=ignore_warnings,
        **kwargs_for_behavioral_test_kwargs_for_estimator,
    )


def get_estimator_classes_that_are_behaviorally_equivalent_wrt_fakit():
    df = test_estimators_serialization()
    ok_to_test_classes = tuple(df[df['kind'] == 'ok']['cls'].values)
    return ok_to_test_classes


if __name__ == '__main__':
    from collections import Counter
    import pandas as pd
    from py2json.inspire.serializing_sklearn_estimators import all_estimator_classes

    n = len(all_estimator_classes)
    print(f'Testing all {n} estimators...')
    df = test_estimators_serialization()

    c = Counter(df.kind)
    print(pd.Series(c).to_string())

    print(
        f"So {c['ok']}/{n} estimators can be automatically created, "
        f'fitted, json-serialized, and deserialized to an object that is '
        f'behaviorally equivalent to the original '
        f'(as far as predict and transform methods are concerned).'
    )
    print('\nThese are those methods...\n')
    print(*[x.__name__ for x in df.cls], sep='\t')



"""
Functions to automatically make json serializers/deserializers for sklearn estimators
"""
from operator import eq
from functools import partial, wraps

import numpy as np
from py2json.util import mk_func_to_kwargs_from_a_val_for_argname_map

from sklearn.utils import all_estimators
from i2.signatures import Sig

all_estimator_classes = tuple([obj for name, obj in all_estimators()])


def funcs_that_need_args(funcs, func_to_kwargs=None, self_name=None):
    """
    >>> from sklearn.utils import all_estimators
    >>> estimator_classes = [obj for name, obj in all_estimators()]
    >>> estimator_names = list(map(lambda x: x.__name__, funcs_that_need_args(all_estimator_classes)))
    >>> expected = [
    ... 'ClassifierChain', 'ColumnTransformer', 'FeatureUnion', 'GridSearchCV', 'MultiOutputClassifier',
    ... 'MultiOutputRegressor', 'OneVsOneClassifier', 'OneVsRestClassifier', 'OutputCodeClassifier',
    ... 'Pipeline', 'RFE', 'RFECV', 'RandomizedSearchCV', 'RegressorChain', 'SelectFromModel',
    ... 'SparseCoder', 'StackingClassifier', 'StackingRegressor', 'VotingClassifier', 'VotingRegressor']
    >>>
    >>> import sklearn
    >>> if sklearn.__version__ == '0.23.1':
    ...      assert estimator_names == expected
    """
    for func in funcs:
        required_args = set(Sig(func).without_defaults.parameters) - {self_name}
        if func_to_kwargs is not None:
            required_args -= func_to_kwargs(func).keys()
        if required_args:
            yield func


from sklearn.decomposition import PCA, TruncatedSVD
from sklearn.linear_model import ElasticNetCV
from sklearn.preprocessing import Normalizer
from scipy.stats import uniform
from collections import (
    UserDict,
)  # using UserDict instead of dict because want the dict to have a doc

estimator_cls_val_for_argname = {
    'estimator': ElasticNetCV,
    'base_estimator': ElasticNetCV,
    'estimators': (PCA, ElasticNetCV),
    'transformers': [
        ('norm1', Normalizer(norm='l1'), [0, 1]),
        ('norm2', Normalizer(norm='l1'), slice(2, 4)),
    ],
    'transformer_list': [
        ('pca', PCA(n_components=1)),
        ('svd', TruncatedSVD(n_components=2)),
    ],
    'steps': [('PCA', PCA), ('ElasticNetCV', ElasticNetCV)],
    'param_grid': {'kernel': ('linear', 'rbf'), 'C': [1, 10]},
    'param_distributions': dict(C=uniform(loc=0, scale=4), penalty=['l2', 'l1']),
}

estimator_cls_val_for_argname = UserDict(estimator_cls_val_for_argname)
estimator_cls_val_for_argname.__doc__ = '''Valid values for default-less arguments of sklearn estimators.
A dict whose keys are all (but one -- sklearn.decomposition._dict_learning.SparseCoder)
the default-less arguments of all sklearn estimators and 
the values provide a valid default for them.

How did you do it, you ask?

Well, I 
>>> val_for_argname = {}  # a mapping from argname and val I updated at every "iteration"
>>> # make a func_to_kwargs from it:
>>> func_to_kwargs = sse.mk_func_to_kwargs_from_a_val_for_argname_map(val_for_argname)
>>>
>>> # make a filter function for that func_to_kwargs
>>> not_covered_by_our_policy = sse.missing_args_func(func_to_kwargs)
>>>
>>> # get the next class that's not covered
>>> not_covered_cls = next(filter(not_covered_by_our_policy, sse.estimator_classes), None)
>>>
>>> if not_covered_cls is not None:
...     from i2.doc_mint import doctest_string_print
...     doctest_string_print(not_covered_cls)
>>> # then I read that example code, figured out a valid input, and re-iterated.
'''

estimator_cls_to_kwargs = mk_func_to_kwargs_from_a_val_for_argname_map(
    estimator_cls_val_for_argname
)

estimator_classes_without_resolved_kwargs = set(
    funcs_that_need_args(all_estimator_classes, estimator_cls_to_kwargs)
)


def is_behaviorally_equivalent(obj1, obj2, behavior_func, output_comp=eq):
    """Checks if two objects obj1 and obj2 are behaviorally equivalent,
    according to behavior func (a callable that will be called on the objects) and
    output_comp (the function that will be called on the outputs of these calls to
    decide whether they can be considered equivalent.

    You'll notice that `behavior_func` takes no other input than `obj` (or `other_obj`).
    Most behaviors depend on other variables you say? We can hardly say that two objects are equivalent
    if they agree on only one data point you say? Yes, I agree. But `is_behaviorally_equivalent` doesn't rule these out.

    What to do if you want to validate behavioral equivalence like the following?
    ```python
    def many_validation_pts():
        for x, y, z in some_large_set_of_xyz_combinations:
            yield output_comp(behavior_func(x, y, obj, z), behavior_func(x, y, other_obj, z))
    assert(all(many_validation_pts()))
    ```

    Well, instead of loading the `is_behaviorally_equivalent` with `args`, `kwargs`, and `argname_of_obj`
    to be able to acheive the above, we simply ask the user to encompass it all in the provided `behavior_func` itself,
    using standard tools such as `functools.partial` or custom decorators.
    For example:

    ```python
    from functools import partial
    def many_validation_pts():
        for x, y, z in some_large_set_of_xyz_combinations:
            _behavior_func = partial(behavior_func, x=x, y=y, z=z)
            yield is_behaviorally_equivalent(obj, other_obj, _behavior_func, output_comp)
    assert(all(many_validation_pts()))
    ```

    Suggestions:
        output_comp=np.allclose or np.isclose
    """
    return output_comp(behavior_func(obj1), behavior_func(obj2))


def init_params_for_cls(estimator_cls):
    """Gets valid params to initialize an estimator_cls
    :param estimator_cls: A (sklearn estimator) class
    :return: init_params such that estimator_cls(**init_params) is valid
    """
    return estimator_cls_to_kwargs(estimator_cls)
    # raise NotImplementedError("")


def xy_for_learner(learner):
    """Gets valid params to initialize an estimator_cls
    :param learner: A (sklearn estimator) INSTANCE
    :return: (X, y) pair such that learner.fit(X, y) works in a meaningful way
    """
    raise NotImplementedError('')


def xy_and_fitted_model_for_estimator_cls(estimator_cls):
    """"""
    init_params = init_params_for_cls(estimator_cls)
    learner = estimator_cls(**init_params)
    X, y = xy_for_learner(learner)
    return X, y, learner.fit(X, y)  # assuming sklearn style


def compose(func1, func2):
    sig1 = Sig(func1)
    sig2 = Sig(func2)

    @sig1
    def composed_funcs(*args, **kwargs):
        return func2(func1(*args, **kwargs))

    composed_funcs.__return_annotation__ = sig2.return_annotation
    return composed_funcs


import pickle

dflt_mk_alt_model = compose(pickle.dumps, pickle.loads)

from sklearn.datasets import make_blobs

_X, _y = make_blobs(center_box=(0, 20.0))


def dflt_xy_for_learner(learner):
    """Returns some random (but fixed after first import) (X, y) pair, insensitive to the learner"""
    return _X, _y


def behavioral_test_kwargs_for_estimator(
    estimator_cls,
    init_params_for_cls=estimator_cls_to_kwargs,  # Returns valid params to initialize an estimator_cls
    xy_for_learner=dflt_xy_for_learner,  # Returns an (X, y) pair for the learner to fit on
    mk_alt_model=dflt_mk_alt_model,  # Returns an alt object by serializing and deserializing the fitted model
    methods=('predict', 'transform',),  # The methods to use to make behavior_funcs,
    output_comp=np.allclose,
):
    """Generate `is_behaviorally_equivalent` kwargs from `estimator_cls`

    :param estimator_cls:
    :param init_params_for_cls: Returns valid params to initialize an estimator_cls
    :param xy_for_learner: Returns an (X, y) pair for the learner to fit on
    :param mk_alt_obj: Returns an alt object by serializing and deserializing the fitted model
    :param methods: The methods to use to make behavior_funcs
    :return Generator of (obj, alt_obj, behavior_func, output_comp) tuples
    """
    model, X, y = model_X_y(estimator_cls, init_params_for_cls, xy_for_learner)
    alt_model = mk_alt_model(model)
    for method in methods:
        if hasattr(estimator_cls, method):
            method_func = getattr(estimator_cls, method)
            yield dict(
                obj1=model,
                obj2=alt_model,
                behavior_func=partial(method_func, X=X),
                output_comp=output_comp,
            )


def model_X_y(estimator_cls, init_params_for_cls, xy_for_learner):
    estimator_params = init_params_for_cls(estimator_cls)
    learner = estimator_cls(**estimator_params)
    X, y = xy_for_learner(learner)
    model = learner.fit(X, y)
    return model, X, y


from py2json.util import catch_errors


def fitting_x_y_works(learner, X, y):
    try:
        learner.fit(X, y)
        return True
    except Exception as e:
        return False


@catch_errors(on_error=None)
def cls_to_learner_or_none(cls, init_params_for_cls=init_params_for_cls):
    return cls(**init_params_for_cls(cls))


@catch_errors(on_error=None)
def fit_or_return_none(learner, X, y):
    learner.fit(X, y)
    return learner


def fit_learners(learners, X, y, ignore_warnings=True):
    with warnings.catch_warnings():
        if ignore_warnings:
            warnings.simplefilter('ignore')
        for i, learner in enumerate(learners):
            try:
                yield fit_or_return_none(learner, X, y)
            except Exception as e:
                print(f'{i}: {learner.__class__.__name__}: {e}')
                yield None


import warnings


class ReprControlledTuple(tuple):
    def __repr__(self):
        t = super().__repr__()
        n = len(t)
        if n > 100:
            return f'{t[:100]}... ({len(self)} elements)'
        else:
            return t


def model_X_y_gen(
    estimator_classes=None,
    init_params_for_cls=estimator_cls_to_kwargs,  # Returns valid params to initialize an estimator_cls
    xy_for_learner=dflt_xy_for_learner,  # Returns an (X, y) pair for the learner to fit on
    on_error='raise',
):
    if estimator_classes is None:
        estimator_classes = all_estimator_classes
    for cls in estimator_classes:
        try:
            yield model_X_y(cls, init_params_for_cls, xy_for_learner)
        except Exception as e:
            if on_error == 'raise':
                raise
            elif on_error == 'ignore':
                continue
            elif on_error == 'yield_context':
                dict(kind='error_in_model_X_y', cls=cls, error=e)
            else:
                raise ValueError(f'Unrecognized on_error: {on_error}')


def test_estimators(
    estimator_classes=None,
    ignore_warnings=True,
    **kwargs_for_behavioral_test_kwargs_for_estimator,
):
    if estimator_classes is None:
        estimator_classes = all_estimator_classes
    with warnings.catch_warnings():
        if ignore_warnings:
            warnings.simplefilter('ignore')

        for estimator_cls in estimator_classes:
            try:
                for kws in behavioral_test_kwargs_for_estimator(
                    estimator_cls, **kwargs_for_behavioral_test_kwargs_for_estimator,
                ):
                    try:
                        if is_behaviorally_equivalent(**kws):
                            yield dict(kind='ok', cls=estimator_cls)
                        else:
                            yield dict(
                                kind='behaviorally_different',
                                cls=estimator_cls,
                                kws=kws,
                            )
                    except Exception as e:
                        yield dict(
                            kind='error_in_test', cls=estimator_cls, kws=kws, error=e,
                        )
            except Exception as e:
                yield dict(kind='error_making_test_kws', cls=estimator_cls, error=e)


def estimator_test_df(
    estimator_classes=None,
    ignore_warnings=True,
    **kwargs_for_behavioral_test_kwargs_for_estimator,
):
    import pandas as pd

    if estimator_classes is None:
        estimator_classes = all_estimator_classes
    return pd.DataFrame(
        list(
            test_estimators(
                estimator_classes,
                ignore_warnings,
                **kwargs_for_behavioral_test_kwargs_for_estimator,
            )
        )
    )


def get_estimator_classes_that_are_behaviorally_equivalent_wrt_pickle():
    df = estimator_test_df()
    ok_to_test_classes = tuple(df[df['kind'] == 'ok']['cls'].values)
    len(ok_to_test_classes)


from py2json.inspire.example_fakit_use import (
    mk_serializer_and_deserializer_from_instance_and_methods,
)



"""A basic JSON encoder to handle numpy and bytes types

>>> bool_array = np.array([True])
>>> bool_value = bool_array[0]
>>> obj = {'an_array': np.array(['a']), 'an_int64': np.int64(1), 'some_bytes': b'a', 'a_bool': bool_value}
>>> assert dumps(obj)
"""

import base64
import json
from functools import partial

import numpy as np


class OtoJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, np.int64):
            return float(obj)
        if isinstance(obj, bytes):
            return base64.b64encode(obj).decode('utf-8')
        if isinstance(obj, np.bool_):
            return True if np.bool_(True) == obj else False

        return json.JSONEncoder.default(self, obj)


json_dump_partial_kwargs = {
    'allow_nan': False,
    'indent': None,
    'separators': (',', ':'),
    'sort_keys': True,
    'cls': OtoJsonEncoder,
}
dump = partial(json.dump, **json_dump_partial_kwargs)
dumps = partial(json.dumps, **json_dump_partial_kwargs)



"""Tools for serialization and deserialization.

An example with numpy arrays:

>>> a = np.array([range(i * 10, (i + 1) * 10) for i in range(10)])
>>> a_decon_ctor_dict = Ctor.deconstruct(a, validate_conversion=True, output_type=Ctor.CTOR_DICT)
>>> a_decon_jdict = Ctor.deconstruct(a, validate_conversion=True, output_type=Ctor.JSON_DICT)
>>> a_recon_from_ctor_dict = Ctor.construct(a_decon_ctor_dict)
>>> a_recon_from_jdict = Ctor.construct(a_decon_jdict)
>>> assert np.allclose(a, a_recon_from_ctor_dict, a_recon_from_jdict)

Another example (not runnable, since dependent on some third party code

```
from ... .source.audio import PyAudioSourceReader

c = PyAudioSourceReader
c_decon_ctor_dict = Ctor.deconstruct(c, validate_conversion=True, output_type=Ctor.CTOR_DICT)
c_decon_jdict = Ctor.deconstruct(c, validate_conversion=True, output_type=Ctor.JSON_DICT)
c_recon_from_ctor_dict = Ctor.construct(c_decon_ctor_dict)
c_recon_from_jdict = Ctor.construct(c_decon_jdict)
assert c == c_recon_from_jdict == c_recon_from_ctor_dict
```


"""
import functools
import importlib
import inspect
import json
from typing import Callable

import numpy as np
from boltons.iterutils import remap, default_enter
from glom import Literal, glom, Spec


def mk_serializer_and_deserializer(spec, mk_inv_spec=None):
    if not isinstance(spec, Spec):
        spec = Spec(spec)

    def serialize(o):
        return glom(o, spec)

    if mk_inv_spec is None:
        return serialize
    else:

        def deserialize(o):
            return glom(o, mk_inv_spec(o))

        return serialize, deserialize


class classproperty(object):
    """
    Similar to @property decorator except for classes instead of class instances
    >>> class Test:
    ...     @classproperty
    ...     def a(cls):
    ...         return 1
    ...
    ...     @property
    ...     def b(self):
    ...         return 2
    >>> Test.a + 1
    2
    >>> Test.b + 1  # Throws TypeError because Test.b is a "property" object until the class is instantiated.
    Traceback (most recent call last):
        ...
    TypeError: unsupported operand type(s) for +: 'property' and 'int'
    """

    def __init__(self, fget):
        self.fget = fget

    def __get__(self, owner_self, owner_cls):
        return self.fget(owner_cls)


def if_condition_return_action(obj, condition_action_list):
    for condition, action in condition_action_list:
        if condition(obj):
            return action(obj)


def is_ctor_jdict(obj):
    return isinstance(obj, dict) and set(list(obj)) == {
        'module',
        'name',
        'attr',
    }


ctor_to_jdict_spec = {
    'description': 'Serialize and deserialize Ctor.CONSTRUCTOR to and from jdict',
    'spec': {
        'module': functools.partial(
            if_condition_return_action,
            condition_action_list=[
                (is_ctor_jdict, lambda obj: obj['module']),
                (inspect.ismethod, lambda obj: obj.__self__.__module__),
                (inspect.isclass, lambda obj: obj.__module__),
                (callable, lambda obj: obj.__module__),
            ],
        ),
        'name': functools.partial(
            if_condition_return_action,
            condition_action_list=[
                (is_ctor_jdict, lambda obj: obj['name']),
                (inspect.ismethod, lambda obj: obj.__self__.__name__),
                (inspect.isclass, lambda obj: obj.__name__),
                (callable, lambda obj: obj.__name__),
            ],
        ),
        'attr': functools.partial(
            if_condition_return_action,
            condition_action_list=[
                (is_ctor_jdict, lambda obj: obj['attr']),
                (inspect.ismethod, lambda obj: obj.__name__),
                (inspect.isclass, lambda obj: None),
                (callable, lambda obj: None),
            ],
        ),
    },
    'mk_inv_spec': lambda d: (
        lambda x: importlib.import_module(d['module']),
        lambda x: getattr(x, d['name']),
        lambda x: getattr(x, d['attr']) if d['attr'] else x,
    ),
}

(
    _serialize_ctor_to_jdict,
    _deserialize_ctor_from_jdict,
) = mk_serializer_and_deserializer(
    ctor_to_jdict_spec['spec'], ctor_to_jdict_spec['mk_inv_spec']
)


class CtorException(ValueError):
    pass


class CtorNames:
    # ctor_dict keys
    CONSTRUCTOR = 'CONSTRUCTOR'  # callable
    ARGS = 'ARGS'  # list
    KWARGS = 'KWARGS'  # dict

    # output types
    CTOR_DICT = 'ctor_dict'
    JSON_DICT = 'jdict'
    CONSTRUCTED = 'constructed'


class Ctor(CtorNames):
    """
    A Base class for serializing any object
    """

    @classmethod
    def _serialize_ctor_to_jdict(cls, ctor_obj):
        return _serialize_ctor_to_jdict(ctor_obj)

    @classmethod
    def _deserialize_ctor_from_jdict(cls, ctor_jdict):
        return _deserialize_ctor_from_jdict(ctor_jdict)

    @classproperty
    def deconstruction_specs(cls):
        """
        Return a list of deconstruction_specs=dict(description='doc string',
                                                   check_type='callable for checking if obj satisfies spec',
                                                   spec='describes how to generate ctor_dict',
                                                   validate_conversion='optional callable comparing obj to ctor_dict')
        """
        return [
            {
                'description': 'For numpy.ndarray',
                'check_type': lambda x: isinstance(x, np.ndarray),
                'spec': {
                    Ctor.CONSTRUCTOR: Literal(np.array),
                    Ctor.ARGS: lambda x: [x.tolist()],
                    Ctor.KWARGS: Literal(None),
                },
                'validate_conversion': lambda x, serialized_x: np.allclose(
                    x, Ctor._construct_obj(serialized_x)
                ),
            },
            {
                'description': 'For objects that have from_jdict and to_jdict methods',
                'check_type': lambda x: hasattr(x, 'from_jdict')
                and hasattr(x, 'to_jdict'),
                'spec': {
                    Ctor.CONSTRUCTOR: lambda x: type(x).from_jdict,
                    Ctor.ARGS: lambda x: [x.to_jdict()],
                    Ctor.KWARGS: Literal(None),
                },
                'validate_conversion': lambda x, serialized_x: x.to_jdict()
                == Ctor._construct_obj(serialized_x).to_jdict(),
            },
            {
                'description': 'for class methods or class constructors',
                'check_type': lambda x: inspect.ismethod(x) or inspect.isclass(x),
                'spec': {
                    Ctor.CONSTRUCTOR: Literal(cls._deserialize_ctor_from_jdict),
                    Ctor.ARGS: lambda x: [cls._serialize_ctor_to_jdict(x)],
                    Ctor.KWARGS: Literal(None),
                },
                'validate_conversion': lambda x, serialized_x: np.allclose(
                    x, Ctor._construct_obj(serialized_x)
                ),
            },
        ]

    @classmethod
    def construct(cls, obj):
        """Construct ctor objs. Return obj with no action if obj is str, int, float, bool, or None.

        :param obj: ctor_dict or jdict
        :return: remap obj with constructed ctor_dicts
        """
        if cls.is_supported_by_default_remap(obj) and cls.is_jdict(obj):
            obj = cls.deserializer(obj)

        if cls.is_ctor_dict(obj):
            return cls._construct_obj(obj)

        if cls.is_supported_by_default_remap(obj):

            def visit(path, key, value):
                if cls.is_ctor_dict(value):
                    value = cls._construct_obj(value)
                return key, value

            def enter(path, key, value):
                if cls.is_ctor_dict(value):
                    return None, False
                else:
                    return default_enter(path, key, value)

            return remap(obj, visit, enter=enter)
        elif cls.is_python_to_json_basic_type(obj):
            return obj
        else:
            CtorException(f'Cannot construct unknown obj type: {type(obj)}')

    @classmethod
    def deconstruct(
        cls,
        obj,
        validate_conversion: bool = False,
        output_type: str = CtorNames.JSON_DICT,
    ):
        """
        Remap a nested structure by deconstructing nodes of non-basic types into ctor_dicts using the deconstruction_specs

        :param obj: any object containing only basic types and those described in the deconstruction_specs
        :param validate_conversion: [boolean] True to compare obj to reconstructed ctor_dict. False to skip validation
        :param output_type: Ctor.JSON_DICT or Ctor.CTOR_DICT
        :return: remapped obj
        """
        decon_ouputs = (Ctor.JSON_DICT, Ctor.CTOR_DICT)
        if output_type not in decon_ouputs:
            raise CtorException(
                f'deconstruct output type must be one of: {decon_ouputs}'
            )

        if not cls.is_supported_by_default_remap(obj):
            if cls.is_python_to_json_basic_type(obj):
                return obj

            ctor_dict = cls._deconstruct_obj(obj)
            if output_type == Ctor.CTOR_DICT:
                return ctor_dict
            elif output_type == Ctor.JSON_DICT:
                return cls.serializer(ctor_dict)

        def visit(path, key, value):
            if not cls.is_python_to_json_basic_type(value) and key != Ctor.CONSTRUCTOR:
                try:
                    value = cls._deconstruct_obj(value, validate_conversion)
                except CtorException:
                    raise CtorException(
                        f'deconstruction spec not found for non-basic type at path={path}, key={key}'
                    )
            return key, value

        def enter(path, key, value):
            if not cls.is_python_to_json_basic_type(value) and key != Ctor.CONSTRUCTOR:
                return None, False
            else:
                return default_enter(path, key, value)

        ctor_dict = remap(obj, visit, enter=enter)
        if output_type == Ctor.CTOR_DICT:
            return ctor_dict
        elif output_type == Ctor.JSON_DICT:
            return cls.serializer(ctor_dict)

    @classmethod
    def serializer(cls, ctor_pydict):
        """
        Search ctor_pydict and replace any Ctor.CONSTRUCTOR nodes with a JSON equivalent
        :param ctor_pydict: A ctor_dict or Dict containing nested ctor_dicts
        :return: jsonizeable ctor_jdict
        """
        return remap(
            ctor_pydict,
            cls._remap_visit_serialize_to_jdict,
            cls._remap_enter_constructor,
        )

    @classmethod
    def deserializer(cls, ctor_jdict):
        """
        Search ctor_jdict and replace any Ctor.CONSTRUCTOR nodes with a module object callable

        :param ctor_jdict: A jsonized ctor_dict or Dict containing nested jsonized ctor_dicts
        :return:
        """
        return remap(
            ctor_jdict,
            cls._remap_visit_deserialize_to_pydict,
            cls._remap_enter_constructor,
        )

    @classmethod
    def to_jdict(cls, ctor_dict):
        return cls.serializer(ctor_dict)

    @classmethod
    def from_jdict(cls, jdict):
        ctor_dict = cls.deserializer(jdict)
        return cls._construct_obj(ctor_dict)

    @classmethod
    def to_ctor_dict(cls, constructor: Callable, args: list, kwargs: dict):
        return {
            cls.CONSTRUCTOR: constructor,
            cls.KWARGS: kwargs,
            cls.ARGS: args,
        }

    # Boolean ..........................................................................................................
    @classmethod
    def is_python_to_json_basic_type(cls, obj):
        """
        Check if JSONEncoder supports the object type by default.

        :param obj: any
        :return: boolean
        """
        return isinstance(obj, (dict, list, tuple, str, int, float, bool, type(None)))

    @classmethod
    def is_ctor_dict(cls, ctor_dict):
        return (
            isinstance(ctor_dict, dict)
            and Ctor.CONSTRUCTOR in ctor_dict
            and callable(ctor_dict[cls.CONSTRUCTOR])
        )

    @classmethod
    def is_supported_by_default_remap(cls, obj):
        return isinstance(obj, (list, tuple, dict, set))

    @classmethod
    def is_jdict(cls, obj):
        try:
            json.dumps(obj)
            is_jdict = True
        except TypeError:
            is_jdict = False
        return is_jdict

    # Private ..........................................................................................................
    @classmethod
    def _construct_obj(cls, ctor_dict):
        """
        Calls the Ctor.CONSTRUCTOR with given Ctor.ARGS and Ctor.KWARGS.
        Checks if ARGS or KWARGS are ctor_dicts and constructs those first

        :param ctor_dict: {Ctor.CONSTRUCTOR: Callable, Ctor.ARGS: List[Any], Ctor.KWARGS: Dict[str, Any]}
        :return: ctor_dict[Ctor.CONSTRUCTOR](*ctor_dict[Ctor.ARGS], **ctor_dict[Ctor.KWARGS])
        """
        if cls.is_ctor_dict(ctor_dict):
            if args := cls._get_value(ctor_dict, cls.ARGS, []):
                args = cls.construct(args)
            if kwargs := cls._get_value(ctor_dict, cls.KWARGS, {}):
                kwargs = cls.construct(kwargs)

            return ctor_dict[cls.CONSTRUCTOR](*args, **kwargs)
        else:
            raise CtorException(
                f"ctor_dict must have a '{cls.CONSTRUCTOR}' key and optionally '{cls.ARGS}' and '{cls.KWARGS}' but got '{ctor_dict}'"
            )

    @classmethod
    def _deconstruct_obj(cls, obj, validate_conversion: bool = False):
        """
        Breakdown an obj into a ctor_dict as described by deconstruction_specs.
        Further breakdowns on deconstructed ARGS and KWARGS if necessary.

        :param obj: any object satisfying one of the deconstruction_specs
        :param validate_conversion: [boolean] True to compare obj to reconstructed ctor_dict. False to skip validation
        :return: ctor_dict: {Ctor.CONSTRUCTOR: Callable, Ctor.ARGS: List[Any], Ctor.KWARGS: Dict[str, Any]}
        """
        try:
            s = next(s for s in cls.deconstruction_specs if s['check_type'](obj))
            try:
                serializer = s['serializer']
            except KeyError:
                s['serializer'] = mk_serializer_and_deserializer(s['spec'])
                serializer = s['serializer']
            ctor_dict = serializer(obj)

            if ctor_dict[Ctor.ARGS] is not None:
                ctor_dict[Ctor.ARGS] = cls.deconstruct(
                    ctor_dict[Ctor.ARGS], validate_conversion=validate_conversion
                )
            if ctor_dict[Ctor.KWARGS] is not None:
                ctor_dict[Ctor.KWARGS] = cls.deconstruct(
                    ctor_dict[Ctor.KWARGS], validate_conversion=validate_conversion
                )

            if validate_conversion is True:
                try:
                    validator = s['validate_conversion']
                except KeyError:
                    s['validate_conversion'] = cls._default_conversion_validation
                    validator = s['validate_conversion']
                if validator(obj, ctor_dict) is False:
                    raise CtorException(
                        f'deconstruction validation test failed for: {obj}'
                    )

            return ctor_dict
        except (StopIteration, KeyError):
            raise CtorException(
                f'deconstruction spec not found for non-basic type: {obj}'
            )

    @classmethod
    def _get_value(cls, dict_obj, key, default_value):
        if key in dict_obj and dict_obj[key] is not None:
            return dict_obj[key]
        else:
            return default_value

    @classmethod
    def _remap_enter_constructor(cls, p, k, v):
        """Determines which nodes to visit. The default_enter points to leaf nodes.
        This additionally looks for Ctor.CONSTRUCTOR nodes."""
        if k == Ctor.CONSTRUCTOR:
            return p, False
        return default_enter(p, k, v)

    @classmethod
    def _remap_visit_serialize_to_jdict(cls, p, k, v):
        """Serialize when visiting Ctor.CONSTRUCTOR node, otherwise keep the same value"""
        if k == Ctor.CONSTRUCTOR:
            return k, Ctor._serialize_ctor_to_jdict(v)
        return k, v

    @classmethod
    def _remap_visit_deserialize_to_pydict(cls, p, k, v):
        """Deserialize when visiting Ctor.CONSTRUCTOR node, otherwise keep the same value"""
        if k == Ctor.CONSTRUCTOR:
            return k, Ctor._deserialize_ctor_from_jdict(v)
        return k, v

    @classmethod
    def _default_conversion_validation(cls, original, ctor_dict):
        return original == Ctor._construct_obj(ctor_dict)



"""
f
"""
from __future__ import division

import re
from copy import copy


def kind_of_type(obj_type):
    return obj_type.__module__ + '.' + obj_type.__name__


def kind_of_obj(obj):
    return kind_of_type(type(obj))


def no_dunder_filt(attr):
    return not attr.startswith('__')


def items_with_transformed_keys(d, key_trans=lambda x: x, key_cond=lambda: True):
    """A generator of (transformed_key, value) items.

    :param d: dict (or mapping) to operate on
    :param key_trans: Function that is applied to the key
    :param key_cond: Function specifying whether to change the key or not
    :return: A generator of (transformed_key, value)

    >>> d = {'a': 1, 2: 20}
    >>> dict(items_with_transformed_keys(d, lambda x: x * 100, lambda x: isinstance(x, int)))
    {'a': 1, 200: 20}

    """
    for k, v in d.items():
        if key_cond(k):
            k = key_trans(k)
        yield k, v


class ApplyDictOf(object):
    pass


apply_dict_of = ApplyDictOf()


class Obj2Dict(object):
    """

    >>> import numpy as np
    >>>
    >>> dc = Obj2Dict(
    ...     to_data_for_kind={
    ...         'numpy.ndarray': lambda obj: obj.tolist()
    ...     },
    ...     from_data_for_kind={
    ...         'numpy.ndarray': lambda data: np.array(data)
    ...     },
    ... )
    >>>
    >>> original_obj = np.array([2,4])
    >>> kind, data = dc.kind_and_data_of_obj(original_obj)
    >>> assert kind == 'numpy.ndarray'
    >>> assert type(data) == list
    >>> assert data == [2, 4]
    >>>
    >>> recovered_obj = dc.obj_of_kind_and_data(kind, data)
    >>> assert type(original_obj) == type(recovered_obj)
    >>> assert all(original_obj == recovered_obj)
    >>>
    >>> # But couldn't make it work (yet) with:
    >>> from collections import Counter
    >>>
    >>> class A(object):
    ...     z = Counter({'n': 10, 'k': 5})
    ...     def __init__(self, x=(1,2,3), y=np.array([2,3,4]), z=None):
    ...         self.x = x
    ...         self._y = y
    ...         if z is not None:
    ...             self.z = z
    ...     def __repr__(self):
    ...         return f"A(x={self.x}, y={self._y}, z={self.z})"
    ...
    >>> dc = Obj2Dict(
    ...     to_data_for_kind={
    ...         'numpy.ndarray': lambda obj: obj.tolist(),
    ...         Counter: dict,
    ...     },
    ...     from_data_for_kind={
    ...         '__main__.A': A,
    ...         'numpy.ndarray': lambda data: np.array(data),
    ...         Counter: Counter
    ...     },
    ... )

    """

    def __init__(self, to_data_for_kind=None, from_data_for_kind=None):
        if to_data_for_kind is None:
            to_data_for_kind = {}
        if from_data_for_kind is None:
            from_data_for_kind = {}

        is_type = lambda x: isinstance(x, type)

        self.to_data_for_kind = dict(
            items_with_transformed_keys(
                to_data_for_kind, key_trans=kind_of_type, key_cond=is_type
            )
        )
        self.from_data_for_kind = dict(
            items_with_transformed_keys(
                from_data_for_kind, key_trans=kind_of_type, key_cond=is_type
            )
        )

    def kind_and_data_of_obj(self, obj):
        kind = kind_of_obj(obj)
        if kind in self.to_data_for_kind:
            return kind, self.to_data_for_kind[kind](obj)
        else:
            return kind, obj

    def obj_of_kind_and_data(self, kind, data):
        if kind.startswith('__builtin__'):
            return data
        if (
            isinstance(data, dict)
            and 'data' in data
            and 'kind' in data
            and len(data) == 2
        ):
            data = self.obj_of_kind_and_data(kind=data['kind'], data=data['data'])

        if kind in self.from_data_for_kind:
            return self.from_data_for_kind[kind](data)
        else:
            return data

    def obj_of_kind_data_dict(self, kind_data_dict):
        return self.obj_of_kind_and_data(
            kind=kind_data_dict['kind'], data=kind_data_dict['data']
        )

    def dict_of(self, obj, attr_filt=no_dunder_filt):
        if attr_filt is None:
            attr_filt = lambda attr: True
        elif isinstance(attr_filt, (list, tuple, set)):
            attr_inclusion_set = set(attr_filt)
            attr_filt = lambda attr: attr in attr_inclusion_set
        elif isinstance(attr_filt, str):
            if attr_filt == 'underscore_suffixed':
                attr_filt = lambda attr: attr.endswith('_')
            else:
                attr_pattern = re.compile(attr_filt)
                attr_filt = attr_pattern.match
        else:
            assert callable(
                attr_filt
            ), "Don't know what to do with that kind of attr_filt: {}".format(attr_filt)

        d = dict()
        for k in filter(attr_filt, dir(obj)):
            attr_obj = getattr(obj, k)
            kind, data = self.kind_and_data_of_obj(attr_obj)
            if data is not apply_dict_of:
                d[k] = {'kind': kind, 'data': data}
            else:
                d[k] = {'kind': kind, 'data': self.dict_of(data, attr_filt)}

        return {'kind': kind_of_obj(obj), 'data': d}

        # def obj_of(self, obj_dict):



"""
A general language for json-serialization of a function call.
- Any construction of a python object needs to go through a function call that makes it so
this approach is general.
- It’s also simple at its base, but open (and intended for) extensions to specialize
and compress the language as well as add layers for security.

Note: "fakit" can be pronounced with the "a" as in "bake" or a
"""
import os
import importlib
from functools import partial
from collections.abc import Mapping

from py2json.util import compose

FAK = '$fak'


# TODO: Make a config_utils.py module to centralize config tools (configs for access is just one
#  -- serializers another)
# TODO: Integrate (external because not standard lib) other safer tools for secrets, such as:
#  https://github.com/SimpleLegal/pocket_protector


def getenv(name, default=None):
    """Like os.getenv, but removes a suffix \\r character if present (problem with some env var
    systems)"""
    v = os.getenv(name, default)
    if v.endswith('\r'):
        return v[:-1]
    else:
        return v


def assert_callable(f: callable) -> callable:
    assert callable(f), f'Is not callable: {f}'
    return f


def dotpath_to_obj(dotpath: str):
    """Loads and returns the object referenced by the string DOTPATH_TO_MODULE.OBJ_NAME"""
    first, *remaining = dotpath.split('.')
    obj = importlib.import_module(first)  # assume it's a module
    for item in remaining:
        obj = getattr(obj, item)
    return obj


def obj_to_dotpath(obj):
    """Get the dotpath reference for an object

    >>> from inspect import Signature
    >>> obj_to_dotpath(Signature.replace)
    'inspect.Signature.replace'
    >>> # note that below, it's not a "full path"
    >>> obj_to_dotpath(func_to_dotpath)[-21:]  # the :21 is because the full string is sys dependent
    'fakit.func_to_dotpath'

    func_to_dotpath is the inverse of dotpath_to_func

    >>> assert dotpath_to_obj(obj_to_dotpath(Signature.replace)) == Signature.replace

    """
    return '.'.join((obj.__module__, obj.__qualname__))


def func_to_dotpath(func: callable) -> str:
    return obj_to_dotpath(assert_callable(func))


def dotpath_to_func(dotpath: str) -> callable:
    """Loads and returns the function referenced by f,
    which could be a callable or a DOTPATH_TO_MODULE.FUNC_NAME dotpath string to one.

    >>> f = dotpath_to_func('os.path.join')
    >>> assert callable(f)  # I got a callable!
    >>> assert f.__name__ == 'join'  # and indeed, it's name is join
    >>> from inspect import signature
    >>> signature(f)  # and it's signature is indeed that of os.path.join:
    <Signature (a, *p)>
    >>>
    >>> # and just for fun...
    >>> assert signature(dotpath_to_func('inspect.signature')) == signature(signature)

    dotpath_to_func is the inverse of func_to_dotpath
    >>> assert dotpath_to_func(func_to_dotpath(signature)) == signature

    """
    obj = dotpath_to_obj(dotpath)
    return assert_callable(obj)


def dflt_func_loader(f) -> callable:
    """Loads and returns the function referenced by f,
    which could be a callable or a DOTPATH_TO_MODULE.FUNC_NAME dotpath string to one,
    or a pipeline of these.
    """
    if callable(f):
        return f
    elif isinstance(f, str):
        return dotpath_to_func(f)
    elif isinstance(f, dict) and len(f) == 1 and FAK in f:
        return fakit(f[FAK])
    else:
        return compose(*map(dflt_func_loader, f))


def _fakit(f: callable, a: (tuple, list), k: dict):
    """The function that actually executes the fak command.
    Simply: `f(*(a or ()), **(k or {}))`

    >>> _fakit(print, ('Hello world!',), {})
    Hello world!

    """
    return f(*(a or ()), **(k or {}))  # slightly protected form of f(*a, **k)


def extract_fak(fak):
    """Extracts the (raw) (f, a, k) triple from a dict or tuple/list fak.
    Also asserts the validity of input fak.

    >>> extract_fak(('func', (1, 2), {'keyword': 3}))
    ('func', (1, 2), {'keyword': 3})

    If fak has only two input items and the second is a dict, the second output will be an empty
    tuple.
    >>> extract_fak(('func', {'keyword': 3}))
    ('func', (), {'keyword': 3})

    If fak has only two input items and the second is a tuple, the second output will be an empty
    dict.
    >>> extract_fak(['func', (1, 2)])
    ('func', (1, 2), {})

    If you only have only one element in your list/tuple input...
    >>> extract_fak(['func'])
    ('func', (), {})

    If your input is a dict
    >>> extract_fak({'f': 'func', 'a': (1, 2), 'k': {'keyword': 3}})
    ('func', (1, 2), {'keyword': 3})
    >>> extract_fak({'f': 'func', 'k': {'keyword': 3}})
    ('func', (), {'keyword': 3})
    >>> extract_fak({'f': 'func', 'a': (1, 2)})
    ('func', (1, 2), {})
    """
    if isinstance(fak, dict):
        if 'f' not in fak:
            raise ValueError(f'There needs to be an `f` key, was not: {fak}')
        f = fak['f']
        a = fak.get('a', ())
        k = fak.get('k', {})
    else:
        assert isinstance(
            fak, (tuple, list)
        ), f'fak should be dict, tuple, or list, was not: {fak}'
        assert (
            len(fak) >= 1
        ), f'fak should have at least one element (the function component): {fak}'
        f = fak[0]
        a = ()
        k = {}
        assert len(fak) in {
            1,
            2,
            3,
        }, 'A tuple fak must be of length 1, 2, or 3. No more, no less.'
        if len(fak) > 1:
            if isinstance(fak[1], dict):
                k = fak[1]
            else:
                a = fak[1]
                assert isinstance(
                    a, (tuple, list)
                ), 'argument specs should be dict, tuple, or list'
            if len(fak) > 2:
                if isinstance(fak[2], dict):
                    assert not k, 'can only have one kwargs'
                    k = fak[2]
                else:
                    assert isinstance(
                        fak[2], (tuple, list)
                    ), 'argument specs should be dict, tuple, or list'
                    assert not a, 'can only have one args'
                    a = fak[2]

    assert isinstance(a, (tuple, list)), f'a kind is not a tuple or list: {fak}'
    assert isinstance(k, Mapping), f'k kind is not a mapping: {fak}'
    return f, a, k


def validate_fak(fak):
    """Returns the input iff (f, a, k) could be extracted and validated from input fak"""
    _ = extract_fak(fak)
    return fak


def is_valid_fak(fak):
    try:
        validate_fak(fak)
        return True
    except Exception:
        return False


def extract_and_load(fak, func_loader=dflt_func_loader):
    f, a, k = extract_fak(fak)
    f = func_loader(f)
    return f, a, k


def fakit(fak, func_loader=dflt_func_loader):
    """Execute a fak with given (f, a, k) tuple or {f: f, a: a, k: k} dict, and a function loader.

    Essentially returns `func_loader(f)(*a, **k)` where `(f, a, k)` are flexibly specified by `fak`.

    The `func_loader` is where you specify any validation of func specification and/or how to get
    a callable from it.
    The default `func_loader` will produce a callable from a dot path (e.g. `'os.path.join'`),
    But note that the intended use is for the user to use their own `func_loader`.
    The user should do this, amongst other things:
    - For security purposes, like not allowing `subprocess.call` or such.
    - For expressivity purposes, like to create their own domain specific mini-language
     that maps function specification to actual function.

    Args:
        fak: A (f, a, k) specification. Could be a tuple or a dict (with 'f', 'a', 'k' keys). All
        but f are optional.
        func_loader: A function returning a function.

    Returns: A python object.

    >>> fak = {'f': 'os.path.join', 'a': ['I', 'am', 'a', 'filepath']}
    >>> assert fakit(fak) =='I/am/a/filepath' or fakit(fak) == 'I\am\a\filepath'

    >>> A = fakit(['collections.namedtuple', ('A', 'x y z')])
    >>> A('no', 'defaults', 'here')
    A(x='no', y='defaults', z='here')

    ... you can also use a dict (which will be understood to be the keyword arguments (`**k`)):

    >>> A = fakit(['collections.namedtuple', {'typename': 'A', 'field_names': 'x y z'}])
    >>> A('no', 'defaults', 'here')
    A(x='no', y='defaults', z='here')

    ... or both:

    >>> A = fakit(['collections.namedtuple', ('A', 'x y z'), {'defaults': ('has', 'defaults')}])
    >>> A('this one')
    A(x='this one', y='has', z='defaults')

    >>> from inspect import signature
    >>>
    >>> A = fakit({'f': 'collections.namedtuple', 'a': ['A', 'x y z'], 'k': {'defaults': (2, 3)}})
    >>> # A should be equivalent to `collections.namedtuple('A', 'x y z', defaults=(2, 3))`
    >>> signature(A)
    <Signature (x, y=2, z=3)>
    >>> A(1)
    A(x=1, y=2, z=3)
    >>> A(42, z='forty two')
    A(x=42, y=2, z='forty two')

    >>> def foo(x, y, z=3):
    ...     return x + y * z
    >>> func_map = {
    ...     'foo': foo,
    ...     'bar': (lambda a, b='world': f"{a} {b}!"),
    ...     'sig': signature}
    >>> from functools import partial
    >>> call_func = partial(fakit, func_loader=func_map.get)
    >>> call_func({'f': 'foo', 'a': (1, 10)})
    31

    Common gotcha: Forgetting that `a` is iterpreted as an iterable of function args. For example

    >>> fakit(('builtins.print', 'hello'))  # not correct
    Traceback (most recent call last):
      ...
    AssertionError: argument specs should be dict, tuple, or list
    >>> fakit(('builtins.print', ['hello']))  # correct
    hello

    >>> fakit(('builtins.sum', [1, 2, 3]))  # not correct
    Traceback (most recent call last):
      ...
    TypeError: sum() takes at most 2 arguments (3 given)
    >>> fakit(('builtins.sum', ([1, 2, 3],)))  # correct
    6

    """
    if isinstance(fak, dict) and FAK in fak:
        fak = fak[FAK]
    return _fakit(*extract_and_load(fak, func_loader))


fakit.w_func_loader = lambda func_loader: partial(fakit, func_loader=func_loader)


def fakit_if_marked_for_it(x, func_loader=dflt_func_loader):
    if isinstance(x, dict) and FAK in x:
        return fakit(x[FAK], func_loader)
    else:
        return x


inf = float('infinity')


def refakit(x, func_loader=dflt_func_loader, max_levels=inf):
    """Fakit recursively looking for nested {'$fak': ...} specifications of python objects

    :param x:
    :param func_loader:
    :return:

    >>> t = {'$fak': ('builtins.sum', ([1,2,3],))}
    >>> refakit(t)  # it works with one level
    6

    >>> ttt = {'$fak': ('builtins.sum', ([t, t],))}
    >>> refakit(ttt)
    12

    But this recursive interpretation of the the fakit elemnts in [t, t] would not
    happen if we restricted the max_levels to be 2 for example.

    The max levels is there to be able to specify that the refakit shouldn't go too deep in
    nested lists (and thus spare some computation.
    TODO: Perhaps we could include this max_levels as a specification in fakit?

    See also: `fakit`, the one level only version of `refakit`.
    """
    if max_levels == 0:
        return x
    if isinstance(x, dict) and FAK in x:
        f, a, k = extract_fak(x[FAK])

        # recurse over inputs to see if there's some that are expressed with a $fak dict
        a = [refakit(aa, func_loader, max_levels - 1) for aa in a]
        k = {kk: refakit(vv, func_loader, max_levels - 1) for kk, vv in k.items()}

        return fakit((f, a, k), func_loader)
    elif isinstance(x, list):
        return [refakit(xx, func_loader, max_levels - 1) for xx in x]
    else:
        return x
